#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <math.h>
#include <limits.h>

#define SETUP "setup_DeltaRule.txt"
#define DELKA_BUFFERU 16384
#define MAX_DELKA_JMENA 128
#define SEPARATORY " ;\t\n"
#define POCET_DES_MIST 15

void cti_string_ze_souboru(FILE *fr, int max_delka, char *s, char *s_bez_pripony)
{
  int i = 0;
  char c, *p_tecka;
  while ((c = getc(fr)) && !strchr(SEPARATORY, c) && i < max_delka - 1) {
    s[i] = c;
    i++;
  }
  s[i] = '\0';
  strcpy(s_bez_pripony, s);
  p_tecka = strrchr(s_bez_pripony, '.');
  if (p_tecka) {
    s_bez_pripony[p_tecka - s_bez_pripony] = '\0';
  }
  printf("Precteno %s%c", s, c);
  while (((c = getc(fr)) != '\n') && !feof(fr))
    putchar(c);
  putchar('\n');
}

int cti_int_ze_souboru(FILE *fr)
{
  int c;
  int cislo;
  if (fscanf(fr, "%d", &cislo) != 1)
    printf ("Chybi hodnota v souboru!\n");
  printf("Precteno %d", cislo);
  while (((c = getc(fr)) != '\n') && !feof(fr))
    putchar(c);
  putchar('\n');
  return cislo;
}

double cti_double_ze_souboru(FILE *fr)
{
  int c;
  double cislo;
  if (fscanf(fr, "%lf", &cislo) != 1)
    printf ("Chybi hodnota v souboru!\n");
  printf("Precteno %g", cislo);
  while (((c = getc(fr)) != '\n') && !feof(fr))
    putchar(c);
  putchar('\n');
  return cislo;
}

int preved_retezec_na_double(char *jmeno_souboru, int r, int s, char *cislo_string, double *cislo)
{
  char *odpad;
  errno = 0;
  *cislo = strtod(cislo_string, &odpad);
  if (errno != 0 || cislo_string == odpad) {
    printf("V souboru %s je na radku %d ve sloupci %d neplatne cislo %s!\t%s\n", jmeno_souboru, r + 1, s + 1, cislo_string, sys_errlist[errno]);
    return 1;
  }
  if (*odpad != '\0') {
    printf("Zahozeno \"%s\" na radku %d ve sloupci %d!\n", odpad, r + 1, s + 1);
  }
  return 0;
}

int nacti(char *jmeno_souboru, int *r, int *s, double ***x, double **p_i) {
  FILE *fr;
  bool nalezen_radek_s_daty = false;
  char buffer_radku[DELKA_BUFFERU];
  char *cislo_string;
  int pocet_prvku_matice = 0, i, j;
  if ((fr = fopen(jmeno_souboru, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    return 1;
  }
  while (fgets(buffer_radku, DELKA_BUFFERU, fr) != NULL) {
    cislo_string = strtok(buffer_radku, SEPARATORY);
    if (cislo_string != NULL) {
      (*r)++;
    }
  }
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    return 1;
  }
  if (!(*r)) {
    printf("Soubor %s nema ani jeden radek s daty.\n", jmeno_souboru);
    return 1;
  }
  if ((fr = fopen(jmeno_souboru, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    return 1;
  }
  while ((fgets(buffer_radku, DELKA_BUFFERU, fr) != NULL) && !nalezen_radek_s_daty) {
    cislo_string = strtok(buffer_radku, SEPARATORY);
    if (cislo_string != NULL) {
      (*s)++;
      nalezen_radek_s_daty = true;
    }
    while ((cislo_string = strtok(NULL, SEPARATORY)) != NULL) {
      (*s)++;
    }
  }
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    return 1;
  }
  if ((fr = fopen(jmeno_souboru, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    return 1;
  }
  while (fgets(buffer_radku, DELKA_BUFFERU, fr) != NULL) {
    cislo_string = strtok(buffer_radku, SEPARATORY);
    if (cislo_string != NULL) {
      pocet_prvku_matice++;
    }
    while ((cislo_string = strtok(NULL, SEPARATORY)) != NULL) {
      pocet_prvku_matice++;
    }
  }
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    return 1;
  }
  if (pocet_prvku_matice != (*r) * (*s)) {
    printf("V souboru %s neni pocet prvku matice (%d) rovny poctu prvku na prvnim radku s daty (%d) krat pocet radku s daty (%d).\n", jmeno_souboru, pocet_prvku_matice, *s, *r);
    return 1;
  }
  if ((*p_i = (double *) malloc((*r) * (*s) * sizeof(double))) == NULL) {
    printf("Malo pameti!\n");
    return 1;
  }
  if ((*x = (double **) malloc((*r) * sizeof(double *))) == NULL) {
    printf("Malo pameti!\n");
    return 1;
  }
  for (i = 0; i < (*r); i++) {
    (*x)[i] = *p_i + i * (*s);
  }
  if ((fr = fopen(jmeno_souboru, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    return 1;
  }
  i = 0; j = 0;
  while (fgets(buffer_radku, DELKA_BUFFERU, fr) != NULL) {
    cislo_string = strtok(buffer_radku, SEPARATORY);
    if (cislo_string != NULL) {
      if (preved_retezec_na_double(jmeno_souboru, i, j, cislo_string, &(*x)[i][j])) {
        return 1;
      }
      j++;
      while ((cislo_string = strtok(NULL, SEPARATORY)) != NULL) {
        if (preved_retezec_na_double(jmeno_souboru, i, j, cislo_string, &(*x)[i][j])) {
          return 1;
        }
        j++;
      }
      i++; j = 0;
    }
  }
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    return 1;
  }
  return 0;
}

void zapis(char *jmeno_souboru, int r, int s, double **x) {
  FILE *fw;
  int i, j;
  if ((fw = fopen(jmeno_souboru, "w")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    exit(1);
  }
  for (i = 0; i < r; i++) {
    for (j = 0; j < s; j++) {
      fprintf(fw, "%.*g\t", POCET_DES_MIST, x[i][j]);
    }
    fprintf(fw, "\n");
  }
  if (fclose(fw) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    exit(1);
  }
}

void uvolni_pamet_pro_matici(double ***x, double **p_i) {
  free((void *) *p_i);
  *p_i = NULL;
  free((void *) *x);
  *x = NULL;
}

void uvolni_pamet_pro_vektor(double **x) {
  free((void *) *x);
  *x = NULL;
}

void zacni(int r, int s, double **x, double **spravny_vystup, double **w) {
  const int PRAH = 1;
  int i;
  if ((*spravny_vystup = (double *) malloc(r * sizeof(double))) == NULL) {
    printf("Malo pameti!\n");
    exit(1);
  }
  if ((*w = (double *) calloc(s, sizeof(double))) == NULL) { /* Function calloc() sets weights to zero. */
    printf("Malo pameti!\n");
    exit(1);
  }
  for (i = 0; i < r; i++) {
    (*spravny_vystup)[i] = x[i][s - 1];
    x[i][s - 1] = PRAH;
  }
}

void pracuj(char *jmeno_souboru, int r, int s, double **x, double *spravny_vystup, double *w, double rychlost, double presnost, int funkce, int max_pruchody) {
  FILE *fw;
  int i, j, pruchody = 0;
  double suma, suma_chyb, y, chyba;
  if ((fw = fopen(jmeno_souboru, "w")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    exit(1);
  }
  do {
    for (i = 0; i < r; i++) { /* Training of weights */
      suma = 0;
      for (j = 0; j < s; j++) {
        suma += x[i][j] * w[j];
      }
      if (funkce == 0) { /* Signum */
        y = (suma > 0) ? 1 : 0;
      } else { /* Sigmoidal */
          y = 1 / (1 + exp(-suma));
        }
      chyba = spravny_vystup[i] - y;
      for (j = 0; j < s; j++) {
        w[j] += rychlost * chyba * x[i][j];
      }
    }
    suma_chyb = 0;
    for (i = 0; i < r; i++) { /* Computing of error */
      suma = 0;
      for (j = 0; j < s; j++) {
        suma += x[i][j] * w[j];
      }
      if (funkce == 0) { /* Sign function */
        y = (suma > 0) ? 1 : 0;
      } else { /* Sigmoid function */
          y = 1 / (1 + exp(-suma));
        }
      chyba = spravny_vystup[i] - y;
      suma_chyb += chyba * chyba;
    }
    fprintf(fw, "%.*g\n", POCET_DES_MIST, suma_chyb);
    pruchody++;
  } while (suma_chyb > presnost && pruchody < max_pruchody);
    if (pruchody == max_pruchody) {
      fprintf(fw, "Bylo dosazeno %d pruchodu trenovaci davkou.\n", pruchody);
    } else {
        fprintf(fw, "V %d. pruchodu trenovaci davkou se vahy jiz nezmenily.\n", pruchody);
      }
  fprintf(fw, "Vektor vah je (");
  for (i = 0; i < s; i++) {
    fprintf(fw, "%.*g ", POCET_DES_MIST, w[i]);
  }
  fprintf(fw, ")\n");
  if (fclose(fw) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    exit(1);
  }
}

int main()
{
  FILE *fr;
  bool nenalezen_radek_s_daty;
  int radek = 0, sloupek = 0, aktivacni_funkce, max_pocet_pruchodu;
  double **x, *p_i, *spravny_vystup, *w, parametr_rychlosti_uceni, presnost;
  char jmeno_souboru[MAX_DELKA_JMENA] = { '\0' };
  char jmeno_souboru_bez_pripony[MAX_DELKA_JMENA] = { '\0' };
  if ((fr = fopen(SETUP, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SETUP);
    return 1;
  }
  cti_string_ze_souboru(fr, MAX_DELKA_JMENA, jmeno_souboru, jmeno_souboru_bez_pripony);
  aktivacni_funkce = cti_int_ze_souboru(fr);
  parametr_rychlosti_uceni = cti_double_ze_souboru(fr);
  presnost = cti_double_ze_souboru(fr);
  max_pocet_pruchodu = cti_int_ze_souboru(fr);
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SETUP);
    return 1;
  }
  nenalezen_radek_s_daty = nacti(jmeno_souboru, &radek, &sloupek, &x, &p_i);
  if (nenalezen_radek_s_daty) {
    return nenalezen_radek_s_daty;
  }
  sprintf(jmeno_souboru, "%s_vystup.txt", jmeno_souboru_bez_pripony);
  zapis(jmeno_souboru, radek, sloupek, x);
  zacni(radek, sloupek, x, &spravny_vystup, &w);
  sprintf(jmeno_souboru, "%s_vysledek_DeltaRule.txt", jmeno_souboru_bez_pripony);
  pracuj(jmeno_souboru, radek, sloupek, x, spravny_vystup, w, parametr_rychlosti_uceni, presnost, aktivacni_funkce, max_pocet_pruchodu);
  uvolni_pamet_pro_matici(&x, &p_i);
  uvolni_pamet_pro_vektor(&spravny_vystup);
  uvolni_pamet_pro_vektor(&w);
  return 0;
}
